@@ -12,7 +12,7 @@
 #define SSL_MODULE
 #include "ssl.h"
 
-static char *CVSid = "@(#) $Id: context.c,v 1.15 2002/09/04 22:24:59 iko Exp $";
+static char *CVSid = "@(#) $Id: context.c,v 1.17 2004/08/06 10:21:56 martin Exp $";
 
 /*
  * CALLBACKS
@@ -264,6 +264,35 @@ ssl_Context_set_passwd_cb(ssl_ContextObj *self, PyObject *args)
     return Py_None;
 }
 
+static char ssl_Context_use_certificate_chain_file_doc[] = "\n\
+Load a certificate chain from a file\n\
+\n\
+Arguments: self - The Context object\n\
+           args - The Python argument tuple, should be:\n\
+             certfile - The name of the certificate chain file\n\
+Returns:   None\n\
+";
+static PyObject *
+ssl_Context_use_certificate_chain_file(ssl_ContextObj *self, PyObject *args)
+{
+    char *certfile;
+
+    if (!PyArg_ParseTuple(args, "s:use_certificate_chain_file", &certfile))
+        return NULL;
+
+    if (!SSL_CTX_use_certificate_chain_file(self->ctx, certfile))
+    {
+        exception_from_error_queue();
+        return NULL;
+    }
+    else
+    {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+}
+
+
 static char ssl_Context_use_certificate_file_doc[] = "\n\
 Load a certificate from a file\n\
 \n\
@@ -825,17 +854,18 @@ ssl_Context_set_options(ssl_ContextObj *self, PyObject *args)
 
 
 /*
- * Member methods in the Connection object
+ * Member methods in the Context object
  * ADD_METHOD(name) expands to a correct PyMethodDef declaration
- *   {  'name', (PyCFunction)ssl_Connection_name, METH_VARARGS }
+ *   {  'name', (PyCFunction)ssl_Context_name, METH_VARARGS }
  * for convenience
- * ADD_ALIAS(name,real) creates an "alias" of the ssl_Connection_real
+ * ADD_ALIAS(name,real) creates an "alias" of the ssl_Context_real
  * function with the name 'name'
  */
 #define ADD_METHOD(name) { #name, (PyCFunction)ssl_Context_##name, METH_VARARGS, ssl_Context_##name##_doc }
 static PyMethodDef ssl_Context_methods[] = {
     ADD_METHOD(load_verify_locations),
     ADD_METHOD(set_passwd_cb),
+    ADD_METHOD(use_certificate_chain_file),
     ADD_METHOD(use_certificate_file),
     ADD_METHOD(use_certificate),
     ADD_METHOD(use_privatekey_file),
@@ -888,7 +918,7 @@ ssl_Context_New(int i_method)
             return NULL;
     }
 
-    self = PyObject_New(ssl_ContextObj, &ssl_Context_Type);
+    self = PyObject_GC_New(ssl_ContextObj, &ssl_Context_Type);
     if (self == NULL)
         return (ssl_ContextObj *)PyErr_NoMemory();
 
@@ -913,28 +943,11 @@ ssl_Context_New(int i_method)
                                 SSL_MODE_AUTO_RETRY);
 
     self->tstate = NULL;
+    PyObject_GC_Track((PyObject *)self);
 
     return self;
 }
 
-/*
- * Deallocate the memory used by the Context object
- *
- * Arguments: self - The Context object
- * Returns:   None
- */
-static void
-ssl_Context_dealloc(ssl_ContextObj *self)
-{
-    SSL_CTX_free(self->ctx);
-    Py_DECREF(self->passphrase_callback);
-    Py_DECREF(self->passphrase_userdata);
-    Py_DECREF(self->verify_callback);
-    Py_DECREF(self->info_callback);
-    Py_DECREF(self->app_data);
-    PyObject_Del(self);
-}
-
 /*
  * Find attribute
  *
@@ -949,6 +962,70 @@ ssl_Context_getattr(ssl_ContextObj *self, char *name)
     return Py_FindMethod(ssl_Context_methods, (PyObject *)self, name);
 }
 
+/*
+ * Call the visitproc on all contained objects.
+ *
+ * Arguments: self - The Context object
+ *            visit - Function to call
+ *            arg - Extra argument to visit
+ * Returns:   0 if all goes well, otherwise the return code from the first
+ *            call that gave non-zero result.
+ */
+static int
+ssl_Context_traverse(ssl_ContextObj *self, visitproc visit, void *arg)
+{
+    int ret = 0;
+
+    if (ret == 0 && self->passphrase_callback != NULL)
+        ret = visit((PyObject *)self->passphrase_callback, arg);
+    if (ret == 0 && self->passphrase_userdata != NULL)
+        ret = visit((PyObject *)self->passphrase_userdata, arg);
+    if (ret == 0 && self->verify_callback != NULL)
+        ret = visit((PyObject *)self->verify_callback, arg);
+    if (ret == 0 && self->info_callback != NULL)
+        ret = visit((PyObject *)self->info_callback, arg);
+    if (ret == 0 && self->app_data != NULL)
+        ret = visit(self->app_data, arg);
+    return ret;
+}
+
+/*
+ * Decref all contained objects and zero the pointers.
+ *
+ * Arguments: self - The Context object
+ * Returns:   Always 0.
+ */
+static int
+ssl_Context_clear(ssl_ContextObj *self)
+{
+    Py_XDECREF(self->passphrase_callback);
+    self->passphrase_callback = NULL;
+    Py_XDECREF(self->passphrase_userdata);
+    self->passphrase_userdata = NULL;
+    Py_XDECREF(self->verify_callback);
+    self->verify_callback = NULL;
+    Py_XDECREF(self->info_callback);
+    self->info_callback = NULL;
+    Py_XDECREF(self->app_data);
+    self->app_data = NULL;
+    return 0;
+}
+
+/*
+ * Deallocate the memory used by the Context object
+ *
+ * Arguments: self - The Context object
+ * Returns:   None
+ */
+static void
+ssl_Context_dealloc(ssl_ContextObj *self)
+{
+    PyObject_GC_UnTrack((PyObject *)self);
+    SSL_CTX_free(self->ctx);
+    ssl_Context_clear(self);
+    PyObject_GC_Del(self);
+}
+
 
 PyTypeObject ssl_Context_Type = {
     PyObject_HEAD_INIT(NULL)
@@ -965,7 +1042,16 @@ PyTypeObject ssl_Context_Type = {
     NULL, /* as_number */
     NULL, /* as_sequence */
     NULL, /* as_mapping */
-    NULL  /* hash */
+    NULL, /* hash */
+    NULL, /* call */
+    NULL, /* str */
+    NULL, /* getattro */
+    NULL, /* setattro */
+    NULL, /* as_buffer */
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
+    NULL, /* doc */
+    (traverseproc)ssl_Context_traverse,
+    (inquiry)ssl_Context_clear,
 };
 
 
